import logging
import pickle
import warnings

from apscheduler import events
from apscheduler.events import JobExecutionEvent, JobSubmissionEvent
from apscheduler.job import Job
from apscheduler.jobstores.base import BaseJobStore, ConflictingIdError, JobLookupError
from apscheduler.schedulers.base import BaseScheduler

from django.core.exceptions import ObjectDoesNotExist
from django.db import connections
from django.db.utils import OperationalError, ProgrammingError

from django_apscheduler.models import DjangoJob
from django_apscheduler.result_storage import DjangoResultStorage
from django_apscheduler.util import deserialize_dt, serialize_dt

LOGGER = logging.getLogger("django_apscheduler")

def ignore_database_error(func):
    from functools import wraps

    @wraps(func)
    def inner(*a, **k):
        try:
            return func(*a, **k)
        except (OperationalError, ProgrammingError) as e:
            warnings.warn(
                "Got OperationalError: {}. "
                "Please, check that you have migrated the database via python manage.py migrate".format(e),
                category=RuntimeWarning,
                stacklevel=3
            )

    return inner

class DjangoJobStore(BaseJobStore):
    """
    Stores jobs in a Django database.
    :param int pickle_protocol: pickle protocol level to use (for serialization), defaults to the
        highest available
    """

    def __init__(self, pickle_protocol=pickle.HIGHEST_PROTOCOL):
        super(DjangoJobStore, self).__init__()
        self.pickle_protocol = pickle_protocol

    @ignore_database_error
    def lookup_job(self, job_id):
        logging.debug("Lookup for a job %s", job_id)
        try:
            job_state = DjangoJob.objects.get(name=job_id).job_state
        except DjangoJob.DoesNotExist:
            return None
        r = self._reconstitute_job(job_state) if job_state else None
        logging.debug("Got %s", r)
        return r

    @ignore_database_error
    def get_due_jobs(self, now):
        logging.debug("get_due_jobs for time=%s", now)
        try:
            out = self._get_jobs(next_run_time__lte=serialize_dt(now))
            logging.debug("Got %s", out)
            return out
        except:
            logging.exception("")

    @ignore_database_error
    def get_next_run_time(self):
        try:
            return deserialize_dt(DjangoJob.objects.first().next_run_time)
        except AttributeError:  # no active jobs
            return None

    @ignore_database_error
    def get_all_jobs(self):
        jobs = self._get_jobs()
        self._fix_paused_jobs_sorting(jobs)
        return jobs

    @ignore_database_error
    def add_job(self, job):
        if DjangoJob.objects.filter(
            name=job.id
        ).exists():
            raise ConflictingIdError(job.id)

        DjangoJob.objects.create(
            name=job.id,
            next_run_time=serialize_dt(job.next_run_time),
            job_state=pickle.dumps(job.__getstate__(), self.pickle_protocol)
        )

    @ignore_database_error
    def update_job(self, job):
        updated = DjangoJob.objects.filter(name=job.id).update(
            next_run_time=serialize_dt(job.next_run_time),
            job_state=pickle.dumps(job.__getstate__(), self.pickle_protocol)
        )

        logging.debug(
            "Update job %s: next_run_time=%s, job_state=%s",
            job,
            serialize_dt(job.next_run_time),
            job.__getstate__()

        )

        if updated == 0:
            logging.info("Job with id %s not found", job.id)
            raise JobLookupError(job.id)

    @ignore_database_error
    def remove_job(self, job_id):
        deleted, _ = DjangoJob.objects.filter(name=job_id).delete()
        if deleted == 0:
            raise JobLookupError(job_id)

    @ignore_database_error
    def remove_all_jobs(self):
        with connections["default"].cursor() as c:
            c.execute("""
                DELETE FROM django_apscheduler_djangojobexecution;
            """)
            c.execute("DELETE FROM django_apscheduler_djangojob;")

    def _reconstitute_job(self, job_state):
        job_state = pickle.loads(job_state)
        job_state['jobstore'] = self
        job = Job.__new__(Job)
        job.__setstate__(job_state)
        job._scheduler = self._scheduler
        job._jobstore_alias = self._alias
        return job

    def _get_jobs(self, **filters):
        job_states = DjangoJob.objects.filter(**filters).values_list('id', 'job_state')
        jobs = []
        failed_job_ids = set()
        for job_id, job_state in job_states:
            try:
                jobs.append(self._reconstitute_job(job_state))
            except:
                self._logger.exception('Unable to restore job "%s" -- removing it', job_id)
                failed_job_ids.add(job_id)

        # Remove all the jobs we failed to restore
        DjangoJob.objects.filter(name__in=failed_job_ids).delete()

        def map_jobs(job):
            job.next_run_time = deserialize_dt(job.next_run_time)
            return job

        return list(map(map_jobs, jobs))


def event_name(code):
    for key in dir(events):
        if getattr(events, key) == code:
            return key


class _EventManager(object):

    LOGGER = LOGGER.getChild("events")

    def __init__(self, storage=None):
        self.storage = storage or DjangoResultStorage()

    def __call__(self, event):
        logging.debug("Got event: %s, %s, %s",
                      event, type(event), event.__dict__)
        # print event, type(event), event.__dict__
        try:
            if isinstance(event, JobSubmissionEvent):
                self._process_submission_event(event)
            elif isinstance(event, JobExecutionEvent):
                self._process_execution_event(event)
        except Exception as e:
            self.LOGGER.exception(str(e))

    def _process_submission_event(self, event):
        # type: (JobSubmissionEvent)->None

        try:
            job = DjangoJob.objects.get(name=event.job_id)
        except ObjectDoesNotExist:
            self.LOGGER.warning("Job with id %s not found in database", event.job_id)
            return

        self.storage.get_or_create_job_execution(job, event)

    def _process_execution_event(self, event):
        # type: (JobExecutionEvent)->None

        try:
            job = DjangoJob.objects.get(name=event.job_id)
        except ObjectDoesNotExist:
            self.LOGGER.warning("Job with id %s not found in database", event.job_id)
            return

        self.storage.register_job_executed(job, event)


def register_events(scheduler, result_storage=None):
    scheduler.add_listener(_EventManager(result_storage))


def register_job(scheduler, *a, **k):
    """

    Helper decorator for job registration.
    Automatically fills id parameter to prevent jobs duplication.
    See this comment for explanation: https://github.com/jarekwg/django-apscheduler/pull/9#issuecomment-342074372

    Usage example::

        @register_job(scheduler, "interval", seconds=1)
        def test_job():
            time.sleep(4)
            print("I'm a test job!")

    :param scheduler: Scheduler instance
    :type scheduler: BaseScheduler

    :param a, k: Params, will be passed to scheduler.add_job method. See :func:`BaseScheduler.add_job`
    """
    # type: (BaseScheduler)->callable

    def inner(func):
        k.setdefault("id", "{}.{}".format(func.__module__, func.__name__))
        scheduler.add_job(func, *a, **k)
        return func

    return inner



import logging
import time

from django_apscheduler.models import DjangoJobExecution
from django_apscheduler.util import serialize_dt


class DjangoResultStorage(object):
    """
    Uses Django ORM table for store job status and results.
    You can override this class to change result storage.
    """
    LOGGER = logging.getLogger("django_apscheduler.result_storage")

    def get_or_create_job_execution(self, job, event):
        """
        Create and return new job execution item.
        :param job: DjangoJob instance
        :param event: JobSubmissionEvent instance
        :return: JobExecution id
        """
        # type: (DjangoJob, JobSubmissionEvent)->int

        # For blocking schedulers we first got FINISH event, and than - SUBMITTED event
        job_execution = DjangoJobExecution.objects.filter(
            job=job,
            run_time=serialize_dt(event.scheduled_run_times[0])
        ).order_by("-id").first()

        if job_execution and job_execution.started is None:
            job_execution.started = time.time()
            job_execution.duration = float(job_execution.finished) - float(job_execution.started)
            job_execution.save()
            return job_execution.id

        return DjangoJobExecution.objects.create(
            job=job,
            status=DjangoJobExecution.SENT,
            started=time.time(),
            run_time=serialize_dt(event.scheduled_run_times[0])
        ).id

    def register_job_executed(self, job, event):
        """
        Registration of job execution status
        :param job: DjangoJob instance
        :param event: JobExecutionEvent instance
        :return: JobExecution id
        """
        # type: (DjangoJobExecution, JobExecutionEvent)->int

        job_execution = DjangoJobExecution.objects.filter(
            job=job,
            status=DjangoJobExecution.SENT,
            run_time=serialize_dt(event.scheduled_run_time)
        ).order_by("id").last()  # type: DjangoJobExecution

        if not job_execution:
            job_execution = DjangoJobExecution.objects.create(
                job=job,
                status=DjangoJobExecution.SENT,
                run_time=serialize_dt(event.scheduled_run_time)
            )

        if job_execution.finished:
            self.LOGGER.warning("Job already finished! %s", job_execution)
            return

        job_execution.finished = time.time()

        try:
            job_execution.duration = float(job_execution.finished) - float(job_execution.started)
        except:
            job_execution.duration = 0

        job_execution.status = DjangoJobExecution.SUCCESS

        if event.exception:
            job_execution.exception = str(event.exception)
            job_execution.traceback = str(event.traceback)
            job_execution.status = DjangoJobExecution.ERROR

        job_execution.save()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-11-05 16:07
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='DjangoJob',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, unique=True)),
                ('next_run_time', models.DateTimeField(db_index=True)),
                ('job_state', models.BinaryField()),
            ],
            options={
                'ordering': ('next_run_time',),
            },
        ),
        migrations.CreateModel(
            name='DjangoJobExecution',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.CharField(choices=[['Added', 'Added'], ['Started execution', 'Started execution'], ['Max instances reached!', 'Max instances reached!'], ['Missed!', 'Missed!'], ['Modified!', 'Modified!'], ['Removed!', 'Removed!'], ['Error!', 'Error!'], ['Executed', 'Executed']], max_length=50)),
                ('run_time', models.DateTimeField(db_index=True)),
                ('duration', models.DecimalField(decimal_places=2, default=None, max_digits=15, null=True)),
                ('started', models.DecimalField(decimal_places=2, default=None, max_digits=15, null=True)),
                ('finished', models.DecimalField(decimal_places=2, default=None, max_digits=15, null=True)),
                ('exception', models.CharField(max_length=1000, null=True)),
                ('traceback', models.TextField(null=True)),
                ('job', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_apscheduler.DjangoJob')),
            ],
            options={
                'ordering': ('-run_time',),
            },
        ),
    ]



from django.conf import settings
from django.utils.timezone import is_aware, is_naive, make_aware, make_naive


def serialize_dt(dt):
    """
    Converts datetime with timezone to datetime without timezone if USE_TZ is False
    :param dt:
    :return:
    """
    if not settings.USE_TZ and is_aware(dt):
        return make_naive(dt)
    return dt


def deserialize_dt(dt):
    if not settings.USE_TZ and is_naive(dt):
        return make_aware(dt)
    return dt



from setuptools import find_packages, setup

setup(
    name='django-apscheduler',
    version='0.2.4',
    description='APScheduler for Django',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Framework :: Django",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Framework :: Django",
        "Framework :: Django :: 1.8",
        "Framework :: Django :: 1.9",
    ],
    keywords='django apscheduler django-apscheduler',
    url='http://github.com/jarekwg/django-apscheduler',
    author='Jarek Glowacki, Stas Kaledin',
    author_email='jarekwg@gmail.com, staskaledin@gmail.com',
    license='MIT',
    packages=find_packages(
        exclude=("tests", )
    ),
    install_requires=[
        'django>=1.8',
        'apscheduler>=3.2.0',
    ],
    zip_safe=False
)



# coding=utf-8
from django.db import models, connection
from django.utils.safestring import mark_safe
import time
import logging

LOGGER = logging.getLogger("django_apscheduler")

class DjangoJobManager(models.Manager):
    """
    This manager pings database each request after 30s IDLE to prevent MysqlGoneAway error
    """
    _last_ping = 0
    _ping_interval = 30

    def get_queryset(self):
        self.__ping()
        return super(DjangoJobManager, self).get_queryset()

    def __ping(self):
        if time.time() - self._last_ping < self._ping_interval:
            return

        try:
            with connection.cursor() as c:
                c.execute("SELECT 1")
        except Exception as e:
            self.__reconnect()

        self._last_ping = time.time()

    def __reconnect(self):
        LOGGER.warning("Mysql closed the connection. Perform reconnect...")
        connection.connection.close()
        connection.connection = None


class DjangoJob(models.Model):
    name = models.CharField(max_length=255, unique=True)  # id of job
    next_run_time = models.DateTimeField(db_index=True)
    # Perhaps consider using PickleField down the track.
    job_state = models.BinaryField()

    objects = DjangoJobManager()

    def __str__(self):
        status = 'next run at: %s' % self.next_run_time if self.next_run_time else 'paused'
        return '%s (%s)' % (self.name, status)

    class Meta:
        ordering = ('next_run_time', )


class DjangoJobExecution(models.Model):
    ADDED = u"Added"
    SENT = u"Started execution"
    MAX_INSTANCES = u"Max instances reached!"
    MISSED = u"Missed!"
    MODIFIED = u"Modified!"
    REMOVED = u"Removed!"
    ERROR = u"Error!"
    SUCCESS = u"Executed"

    job = models.ForeignKey(DjangoJob, on_delete=models.CASCADE)
    status = models.CharField(max_length=50, choices=[
        [x, x]
        for x in [ADDED, SENT, MAX_INSTANCES, MISSED, MODIFIED,
                  REMOVED, ERROR, SUCCESS]
    ])
    run_time = models.DateTimeField(db_index=True)
    duration = models.DecimalField(max_digits=15, decimal_places=2,
                                   default=None, null=True)

    started = models.DecimalField(max_digits=15, decimal_places=2,
                                  default=None, null=True)
    finished = models.DecimalField(max_digits=15, decimal_places=2,
                                   default=None, null=True)

    exception = models.CharField(max_length=1000, null=True)
    traceback = models.TextField(null=True)

    def html_status(self):
        m = {
            self.ADDED: "RoyalBlue",
            self.SENT: "SkyBlue",
            self.MAX_INSTANCES: "yellow",
            self.MISSED: "yellow",
            self.MODIFIED: "yellow",
            self.REMOVED: "red",
            self.ERROR: "red",
            self.SUCCESS: "green"
        }

        return mark_safe("<p style=\"color: {}\">{}</p>".format(
            m[self.status],
            self.status
        ))

    def __unicode__(self):
        return "Execution id={}, status={}, job={}".format(
            self.id, self.status, self.job
        )

    class Meta:
        ordering = ('-run_time', )



import datetime

from django.contrib import admin
from django.db.models import Avg
from django.utils.timezone import now

from django_apscheduler.models import DjangoJob, DjangoJobExecution


def execute_now(ma, r, qs):
    for item in qs:
        item.next_run_time = now()
        item.save()


execute_now.short_description = "Force tasks to execute right now"


@admin.register(DjangoJob)
class DjangoJobAdmin(admin.ModelAdmin):
    list_display = ["id", "name", "next_run_time_sec", "average_duration"]
    actions = []

    def get_queryset(self, request):
        self._durations = {
            item[0]: item[1]
            for item in DjangoJobExecution.objects.filter(
                status=DjangoJobExecution.SUCCESS,
                run_time__gte=now() - datetime.timedelta(days=2)
            ).values_list("job").annotate(duration=Avg("duration"))
        }
        return super(DjangoJobAdmin, self).get_queryset(request)

    def next_run_time_sec(self, obj):
        return obj.next_run_time.strftime("%Y-%m-%d %H:%M:%S")

    def average_duration(self, obj):
        return self._durations.get(obj.id) or 0


@admin.register(DjangoJobExecution)
class DjangoJobExecutionAdmin(admin.ModelAdmin):
    list_display = ["id", "job", "html_status", "run_time_sec", "duration"]

    list_filter = ["job__name", "run_time", "status"]

    def run_time_sec(self, obj):
        return obj.run_time.strftime("%Y-%m-%d %H:%M:%S")

    def get_queryset(self, request):
        return super(DjangoJobExecutionAdmin, self).get_queryset(
            request
        ).select_related("job")
